###########################################################################
# date : Jan 2024
# Implementation of core functions of pGVEM

###########################################################################

# one iteration of Expectation-Maximization
# in exploratory case, the Correlation \Sigma_\theta is fixed as identity
# in one iteration, we first compute the mean and variance of the variational distribution
# next based on the variational distribution, we maximize the lower bound of expectation of the likelihood by updating the item parameters together with the variational parameter xi
# for convenient the input is \eta(\xi_{ijk}^{(t-1)}) and out put is \xi_{ijk}^{(t)} for the calculation of evidence lower bound 
gvem_ite_once <- function(response, a_hat, b_hat, xi_hat, K_list)
{
  I <- dim(response)[1];J <- dim(response)[2];
  D <- dim(a_hat)[2];
  K <- max(K_list)
  mu_hat <- array(0, dim=c(I,D));
  Sig_hat <- array(0, dim=c(I,D,D));
  b_hat <- cbind(0, b_hat);
  hxi <- array(1 / 8,dim=c(I,J,(K + 1)))
  for(i in 1:I)
    for(j in 1:J)
      for(k in 1:(K_list[j] + 1))
        if(k != (response[i,j] + 1)){
          hxi[i,j,k]=quaeta(xi_hat[i,j,k]);}
  #E Step
  #update mean and covariance of the variational distribution
  for(i in 1:I)
  {
    Sig_hat[i,,] <- diag(D);
    muarr <- array(0, dim = c(D));
    for(j in 1:J)
    {
      kij <- response[i,j] + 1;
      for(k in 1:(K_list[j] + 1))
      {
        if(k!=kij)
        {
          dw <- (k-kij) * a_hat[j,];
          x <- hxi[i,j,k]
          Sig_hat[i,,] <- Sig_hat[i,,] + 2 * x * dw %*% t(dw)
          muarr <- muarr + (2 * x * (b_hat[j, k] - b_hat[j, kij]) - 0.5) * dw
        }
      }
    }
    Sig_hat[i,,] <- solve(Sig_hat[i,,]);
    mu_hat[i,] <- Sig_hat[i,,] %*% muarr;
  }
  
  #M Step
  #update the discrimination parameters
  for(j in 1:J)
  {
    wmat <- array(0,dim=c(D,D));warr <- array(0,dim=c(D));
    for(k in 1:(K_list[j] + 1))
    {
      for(i in 1:I){
        kij <- response[i,j]+1;
        if(k!=kij)
        {
          x <- hxi[i, j, k]
          wmat <- wmat + x * (k - kij)^2 * (Sig_hat[i,,] + mu_hat[i,] %*% t(mu_hat[i,]))
          warr <- warr - 0.5 * (k - kij) * mu_hat[i,] + 2 * x * (b_hat[j, k] - b_hat[j, kij]) * (k - kij) * mu_hat[i,]
        }
      }
    }
    
    a_hat[j,] <- 0.5 * solve(wmat,warr);
  }

  #update the difficulty parameters
  for(j in 1:J)
  {
    for(k in 2:(K_list[j] + 1))
    {
      bs1 <- 0;bs2 <- 0;
      for(i in 1:I)
      {
        kij <- response[i,j]+1;
        dw <- a_hat[j,]%*%mu_hat[i,];
        if(k!=kij)
        {
          x <- hxi[i, j, k]
          bs1 <- bs1 + x
          bs2 <- bs2 + 2 * x * (k - kij) * dw + 0.5 + 2 * x * b_hat[j, kij]
        }
        if(k==kij)
        {
          for(v in 1:(K_list[j] + 1))
          {
            if(v!=k)
            {
              x <- hxi[i, j, v]
              bs1 <- bs1 + x
              bs2 <- bs2 - 2 * x * (v - k) * dw - 0.5 + 2 * x * b_hat[j, v]
            }
          }
        }
      }
      b_hat[j,k]=0.5*bs2/bs1;
    }
  }
  #update the variational parameters
  for(i in 1:I)
    for(j in 1:J)
    {
      q <- a_hat[j,]%*%Sig_hat[i,,]%*%a_hat[j,];
      p <- a_hat[j,]%*%mu_hat[i,];
      kij <- response[i, j] + 1;
      for(k in 1:(K_list[j] + 1))
      {
        if(kij!=k)
        {
          dw <- (k - kij) * p
          db <- b_hat[j, k] - b_hat[j, kij]
          vec <- dw - db
          xs <- vec^2 + (k - kij)^2 * q
          xi_hat[i, j, k] <- sqrt(max(0, xs))
        }
        if(kij==k)
          xi_hat[i, j, k] <- 0;
      }
    }
  
  
  return(list(para_a=a_hat, para_b=(b_hat[, 2:(K+1)]), para_mu=mu_hat, para_sigm=Sig_hat, para_xi=xi_hat));
  
}

gvemeva<-function(response, D, maxiter=500, SE=FALSE, B=50){
  
  
  I <- dim(response)[1];
  J <- dim(response)[2];
  K_list=c();
  for(j in 1:J)
    K_list[j] <- max(response[,j]);
  K <- max(K_list);
  
  threshold <- 1e-05;
  
  mu_hat <- array(0,dim = c(I,D));Sig_hat <- array(0,dim=c(I,D,D));
  
  a_hat <- array(0,dim = c(J,D));b_hat <- array(0,dim = c(J,K));xi_hat <- array(0,dim = c(I,J,(K+1)));
  
  for(i in 1:I) Sig_hat[i,,] <- diag(D);
  

  #initialize item parameters
  for(j in 1:J){
    for(d in 1:D)
      a_hat[j,d] <- runif(1,0.5,1);
    for(k in 1:K_list[j])
      b_hat[j,k]=rnorm(1,0,1)
  }
  #initialize variational parameters
  for(i in 1:I)
    for(j in 1:J)
    {
      for(k in 1:(K_list[j] + 1))
        xi_hat[i,j,k] <- rnorm(1,0,1);
      xi_hat[i,j,(response[i, j] + 1)] <- 0;
    }
  
  start_time=Sys.time();
  #iteration stops when the different of item parameters from two conservative falls below the threshold
  oldapara=c(c(a_hat),c(b_hat));
  for(ite in 1:maxiter)
  {
    ite_result <- gvem_ite_once(response, a_hat, b_hat, xi_hat, K_list);
    mu_hat <- ite_result$para_mu;
    Sig_hat <- ite_result$para_sigm;
    a_hat <- ite_result$para_a;
    b_hat <- ite_result$para_b;
    xi_hat <- ite_result$para_xi;
    
    #compute the ca_hatnge of item parameters
    para <- c(c(a_hat),c(b_hat));
    diff <- sum(((oldapara-para)*(oldapara-para))/( J*D + sum(K_list)));
    if(diff<(threshold))
    {break;}
    oldapara <- para;
  }
  end_time <- Sys.time();
  evatime <- end_time-start_time;
  
  #compare the likelihood
  #use the evidence lower bound
  llk <- elbo(I, J, D, K_list, a_hat, b_hat, Sig_hat, mu_hat, xi_hat, response);
  
  #compute the covariance matrix of latent ability after EFA
  sig_hat <- diag(D)
  
  #promax rotation
  if(D>1){
  rot <- (promax(a_hat,m=4))
  
  G <- (rot$rotmat);
  a_hat <- a_hat %*% (G)
  sig_hat <- (solve(G)) %*% sig_hat %*% t(solve(G));
  mu_hat <- mu_hat %*% solve(G);
  }
  se <- sqrt(diag(sig_hat));
  for(d in 1:D)
  {
    sig_hat[,d] <- sig_hat[,d]/se;
    sig_hat[d,] <- sig_hat[d,]/se;
  }
  #
  for(d in 1:D)
  {
    if(mean(a_hat[,d])<0){
      a_hat[,d] <- -a_hat[,d];mu_hat[,d] <- -mu_hat[,d];
      hatsig[,d] <- -hatsig[,d];hatsig[d,] <- -hatsig[d,];}
  }
  if(!SE)
    return(list(a_hat = (a_hat), b_hat = (b_hat), Sig_hat = (sig_hat), thet_hat = (mu_hat), elbo = llk, se_a = c(0), se_b = c(0), time = as.numeric(evatime)));
  
  btp_p <- array(0, dim = c(I, J, (K + 1)))
  
  for (i in 1:I) {
    for (j in 1:J) {
      exparr <- c(1)
      for (k in 1:K_list[j])
        exparr[k] <- exp(k * a_hat[j,] %*% mu_hat[i,] - b_hat[j, k])
      s <- sum(exparr) + 1
      btp_p[i, j, 1] <- 1.0 / s
      for (k in 1:K_list[j])
        btp_p[i, j, (k + 1)] <- exparr[k] / s
    }
  }
  
  ICrow <- c(c(a_hat),c(b_hat));
  
  IC <- array(0,dim=c(B,((K + D) * J)))
  btp=1;
  total_count=1;
  while(btp<B)
  {
    total_count=total_count+1;
    btp_resp <- array(0,dim=c(I,J));
    for(i in 1:I)
      for(j in 1:J)
      {
        x=runif(1, 0, 1)
        for(k in 1:(K_list[j]+1))
          if(x>sum(btp_p[i,j,1:k])) {btp_resp[i,j] <- k;}
      }
    fskip <- ifResp(btp_resp, K_list);
    if(!fskip)
    {
      next;
    }
    btp_est <- gvemeva(btp_resp,D)
    IC[btp,] <- c(c(btp_est$a_hat),c(btp_est$b_hat));
    
    btp <- btp+1;
    if(total_count>200)
      break;
  }
  if(btp<1)
   return(0);
  btp_se=c();
  for(s in 1:((K+D)*J))
  {
    btp_se[s] <- 0;
    for(b in 1:btp)
      btp_se[s] <- btp_se[s]+(IC[b,s]-ICrow[s])^2;
    btp_se[s] <- sqrt(btp_se[s]/(b-1));
  }
  btp_se_a <- matrix(btp_se[1:(J * D)], ncol = D, nrow = J);btp_se_b=matrix(btp_se[(J * D + 1):(J * D + J * K)], ncol = K, nrow = J)
  
  return(list(a_hat = (a_hat), b_hat = (b_hat), Sig_hat = (sig_hat), thet_hat = (mu_hat), elbo = llk, se_a = btp_se_a, se_b = btp_se_b, time = as.numeric(evatime)));
}
